@@ -1,7 +1,7 @@
 /*
   Bruteforce a wallet file.
 
-  Copyright 2014-2017 Guillaume LE VAILLANT
+  Copyright 2014-2024 Guillaume LE VAILLANT
 
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -64,7 +64,6 @@ unsigned int nb_threads = 1;
 unsigned char last_pass[LAST_PASS_MAX_SHOWN_LENGTH];
 time_t start_time;
 unsigned int status_interval = 0;
-struct itimerval progress_timer, state_timer;
 struct decryption_func_locals
 {
   unsigned long long int counter;
@@ -77,7 +76,7 @@ unsigned int *tab = NULL;
  * Statistics
  */
 
-void handle_signal(int signo)
+void print_status()
 {
   unsigned long long int total_ops = 0;
   unsigned int i, l;
@@ -109,7 +108,9 @@ void handle_signal(int signo)
   else
     fprintf(stderr, "Tried passwords: %llu\n", total_ops);
   fprintf(stderr, "Tried passwords per second: %lf\n", pw_per_seconds);
+  pthread_mutex_lock(&get_password_lock);
   fprintf(stderr, "Last tried password: %s\n", last_pass);
+  pthread_mutex_unlock(&get_password_lock);
   if(dictionary == NULL)
   {
     fprintf(stderr, "Total space searched: %lf%%\n", (total_ops / space) * 100);
@@ -129,6 +130,31 @@ void handle_signal(int signo)
   fprintf(stderr, "\n");
 }
 
+void * status_func(void *arg)
+{
+  unsigned int i;
+
+  while(stop == 0)
+  {
+    for(i = 0; i < status_interval; i++)
+    {
+      sleep(1);
+      if(stop == 1)
+      {
+        pthread_exit(NULL);
+      }
+    }
+    print_status();
+  }
+
+  pthread_exit(NULL);
+}
+
+void handle_signal(int signo)
+{
+  print_status();
+}
+
 
 /*
  * Decryption
@@ -356,7 +382,7 @@ void * decryption_func(void *arg)
       if((ret == 1) && valid_seckey(seckey, seckey_len1 + seckey_len2, pubkey, pubkey_len))
       {
         /* We have a positive result */
-        handle_signal(SIGUSR1); /* Print some stats */
+        print_status();
         pthread_mutex_lock(&found_password_lock);
         found_password = 1;
         printf("Password found: %s\n", pwd);
@@ -385,7 +411,7 @@ void * decryption_func(void *arg)
  * Save/restore state
  */
 
-void save_state(int signo)
+void save_state()
 {
   unsigned int i;
   unsigned long long int total_ops = 0;
@@ -426,6 +452,26 @@ void save_state(int signo)
   fclose(state);
 }
 
+void * state_func(void *arg)
+{
+  unsigned int i;
+
+  while(stop == 0)
+  {
+    for(i = 0; i < 60; i++)
+    {
+      sleep(1);
+      if(stop == 1)
+      {
+        pthread_exit(NULL);
+      }
+    }
+    save_state();
+  }
+
+  pthread_exit(NULL);
+}
+
 void restore_state()
 {
   unsigned int i, n;
@@ -712,7 +758,7 @@ void usage(char *progname)
 
 int main(int argc, char **argv)
 {
-  pthread_t *decryption_threads;
+  pthread_t *decryption_threads, status_thread, state_thread;
   int i, ret, c;
 
   setlocale(LC_ALL, "");
@@ -911,12 +957,12 @@ int main(int argc, char **argv)
   signal(SIGUSR1, handle_signal);
   if(status_interval > 0)
   {
-    signal(SIGALRM, handle_signal);
-    progress_timer.it_value.tv_sec = status_interval;
-    progress_timer.it_value.tv_usec = 0;
-    progress_timer.it_interval.tv_sec = status_interval;
-    progress_timer.it_interval.tv_usec = 0;
-    setitimer(ITIMER_REAL, &progress_timer, NULL);
+    ret = pthread_create(&status_thread, NULL, &status_func, NULL);
+    if(ret != 0)
+    {
+      perror("Error: status thread");
+      exit(EXIT_FAILURE);
+    }
   }
 
   pthread_mutex_init(&found_password_lock, NULL);
@@ -936,12 +982,12 @@ int main(int argc, char **argv)
   {
     restore_state();
 
-    signal(SIGVTALRM, save_state);
-    state_timer.it_value.tv_sec = 60 * nb_threads;
-    state_timer.it_value.tv_usec = 0;
-    state_timer.it_interval.tv_sec = 60 * nb_threads;
-    state_timer.it_interval.tv_usec = 0;
-    setitimer(ITIMER_VIRTUAL, &state_timer, NULL);
+    ret = pthread_create(&state_thread, NULL, &state_func, NULL);
+    if(ret != 0)
+    {
+      perror("Error: state thread");
+      exit(EXIT_FAILURE);
+    }
   }
 
   /* Start decryption threads. */
@@ -959,9 +1005,18 @@ int main(int argc, char **argv)
   {
     pthread_join(decryption_threads[i], NULL);
   }
+  stop = 1;
+  if(state_file != NULL)
+  {
+    pthread_join(state_thread, NULL);
+  }
+  if(status_interval > 0)
+  {
+    pthread_join(status_thread, NULL);
+  }
   if(found_password == 0)
   {
-    handle_signal(SIGUSR1); /* Print some stats */
+    print_status();
     fprintf(stderr, "Password not found\n");
   }
   free(thread_locals);
